from setuptools import setup
setup(
    name = "pengines",
    packages = ["pengines"],
    version = "0.1.8",
    description = "A simple python library for interacting with SWI-Prologs Pengines",
    long_description=open("README.rst").read(),
    author = "Ian Andrich, Chris Mungall",
    author_email = "iandrich87+github@gmail.com",
    url = "https://github.com/ian-andrich/PythonPengines",
    download_url = "https://github.com/ian-andrich/PythonPengines/0.1.4",

    keywords = ["pengines", "pengine", "prolog", "swipl", "swi-prolog", "logic programming"],
    classifiers = [],
)
